% ******************************************************************************
% Prototypical AFEM example with iterative solver for higher order finite
% elements steered by ZZ-estimator with Scott-Zhang as averaging operator.
% ******************************************************************************

%% parameters
nDofsMax = 1e5;
theta = 0.5;
pmax = 4;
lamalg = 0.1;

%% initialization for every polynomial degree
[nElem, nDofs, errEst, recovery, oscillation, iterations, time, directSolveTime] = deal(zeros(pmax, 1000));
for p = 1:pmax
    %% setup geometry & spaces
    printLogMessage('*** p = %d (of %d) ***', p, pmax)
    mesh = Mesh.loadFromGeometry('Lshape');
    fes = FeSpace(mesh, HigherOrderH1Fe(p), 'dirichlet', ':');
    u = FeFunction(fes);
    u.setData(0);

    %% set problem data for -\Delta u = 1 on L-shape
    blf = BilinearForm();
    lf = LinearForm();

    blf.a = Constant(mesh, 1);
    lf.f = Constant(mesh, 1);

    %% set up solver and operator for nested iteration
    % choose the iterative solver of choice: multigrid (with the variants
    % lowOrderVcycle and highOrderVcycle), pcg (with jacobi and additive
    % Schwarz preconditioner), and the cg solver
    %[solver, P] = IterativeSolver.choose(fes, blf, "pcg", "iChol");
    %[solver, P] = IterativeSolver.choose(fes, blf, "pcg", "additiveSchwarzHighOrder");
    [solver, P] = IterativeSolver.choose(fes, blf, "multigrid", "lowOrderVcycle");
    solver.tol = 1e-6;
    solver.maxIter = 100;

    %% adaptive loop
    ell = 0;
    meshSufficientlyFine = false;
    while ~meshSufficientlyFine
        %% setup
        tic;
        ell = ell + 1;
        A = assemble(blf, fes);
        rhs = assemble(lf, fes);

        freeDofs = getFreeDofs(fes);
        A = A(freeDofs,freeDofs);
        solver.setupSystemMatrix(A);
        solver.setupRhs(rhs(freeDofs), u.data(freeDofs)');

        % exact solution as reference
        tic;
        x_ex = A \ rhs(freeDofs);
        directSolveTime(p, ell) = toc;
        
        %% iterative solve
        % solver.solve() implements the algebraic solver loop and
        % terminates automatically once the wanted accuracy is reached
        % Another option is to manually implement a loop and implement one
        % solver step with solver.step() and stop by setting a stopping
        % criterion
        solverIsConverged = false;
        while ~solverIsConverged
            solver.step();
            u.setFreeData(solver.x)
            [mu2, recoveryTerm, oscillationTerm] = estimateZZ(blf, lf, u);
            estimator = sqrt(sum(mu2));
            solver.tol = lamalg*estimator;
            solverIsConverged = solver.isConverged;
        end
        

        %% estimate error and store data
        nDofs(p,ell) = getDofs(fes).nDofs;
        errEst(p,ell) = estimator;
        recovery(p,ell) = sqrt(sum(recoveryTerm));
        oscillation(p,ell) = sqrt(sum(oscillationTerm));
        iterations(p, ell) = solver.iterationCount();
        nElem(p,ell) = mesh.nElements;
        printLogMessage('number of dofs: %d, estimator: %.2e after %d iterations', ...
            nDofs(p,ell), errEst(p,ell), solver.iterationCount);

        %% stoping criterion
        meshSufficientlyFine = (nDofs(p,ell) > nDofsMax);

        %% refine mesh
        if ~meshSufficientlyFine
            marked = markDoerflerSorting(mu2, theta);
            mesh.refineLocally(marked, 'NVB');
            u.setData(prolongate(P, u));
        end
        time(p,ell) = toc;
    end
end

%% plot
figure()
colors = [[0      0.4470 0.7410];
          [0.8500 0.3250 0.0980];
          [0.9290 0.6940 0.1250];
          [0.4940 0.1840 0.5560];
          [0.4660 0.6740 0.1880];
          [0.3010 0.7450 0.9330];
          [0.6350 0.0780 0.1840]];
for p = 1:pmax
    idx = find(nDofs(p,:) > 0);
    complexity = [cumsum(nDofs(p, idx) .* iterations(p, idx))];
    loglog(complexity, errEst(p,idx), '-o', 'LineWidth', 2, 'DisplayName', ['\mu for p=',num2str(p)], 'Color', colors(p,:));
    hold on
    loglog(complexity, recovery(p,idx), '-x', 'LineWidth', 2, 'DisplayName', ['recovery for p=',num2str(p)], 'Color', colors(p,:));
end
for p = unique([1,pmax])
    x = complexity / complexity(1);
    loglog(complexity(1)*x, errEst(p,1)*x.^(-p/2), '--', 'LineWidth', 2, 'Color', 'k', 'DisplayName', ['\alpha = ', num2str(p), '/2'])
end
legend
xlabel('computational cost')
ylabel('\eta')
title('error estimator over number of total computational cost')


%% local function for recovery based a posteriori error estimation

function [indicators, recoveryTerm, oscillationTerm] = estimateZZ(~, lf, u)

p = u.fes.finiteElement.order;
mesh =  u.fes.mesh;
trafo = getAffineTransformation(mesh);
qr = QuadratureRule.ofOrder(2*p);

% Remark: Scott-Zhang operator for piecewise polynomials of degree p corresponds to nodal interpolation

% nodal interpolation of first component of gradient of u
gGradU1 = FeFunction(u.fes);
gGradU1.setData(nodalInterpolation(CompositeFunction(@(grad) grad(1,:,:), Gradient(u)), u.fes));

% nodal interpolation of second component of gradient of u
gGradU2 = FeFunction(u.fes);
gGradU2.setData(nodalInterpolation(CompositeFunction(@(grad) grad(2,:,:), Gradient(u)), u.fes));

% compute recovery term
recoveryTermHandle = @(g1, g2, grad) (g1 - grad(1,:,:)).^2 + (g2 - grad(2,:,:)).^2;
recoveryTermFunction = CompositeFunction(recoveryTermHandle, gGradU1, gGradU2, Gradient(u));
recoveryTerm = integrateElement(recoveryTermFunction, qr);

% element-wise integrals of residual and residual squared
R = CompositeFunction(@(u,f) u(1,:,:) + u(4,:,:) + f, Hessian(u), lf.f);
rIntegrals = integrateElement(R, qr);
rSquaredIntegrals = integrateElement(CompositeFunction(@(R) R.^2, R), qr);

% compute patches
patches = computePatches(mesh.elements);
patchAreas = mesh.computePatchAreas;

% compute oscillation term (projection only for p=0, i.e., integral mean)
oscillationTerm = zeros(1, mesh.nElements);
boundaryVertices = mesh.edges(:,mesh.getCombinedBndEdges);
interiorVertices = true(1, mesh.nCoordinates);
interiorVertices(boundaryVertices) = false;
interiorVertices = find(interiorVertices);
for i = 1:length(interiorVertices)
    z = interiorVertices(i);
    rPatchIntegral = sum(rIntegrals(patches{z}));
    rSquaredPatchIntegral = sum(rSquaredIntegrals(patches{z}));
    diff_integral = rSquaredPatchIntegral - (rPatchIntegral)^2 / patchAreas(z);
    oscillationTerm(patches{z}) = oscillationTerm(patches{z}) + diff_integral;
end

% combine terms suitably
hT = sqrt(trafo.area);
oscillationTerm = hT.^2 .* oscillationTerm;
indicators = recoveryTerm + oscillationTerm;
end

%% local function for computing patches for each vertex

function patches = computePatches(elements)
    n = size(elements, 1);
    N = size(elements, 2);

    patchCounts = accumarray(elements(:), 1);
    ptr = cumsum([1; patchCounts]);
    patchElements = zeros(n*N, 1);

    for i = 1:N
        for k = 1:n
            idx = elements(k,i);
            patchElements(ptr(idx)) = i;
            ptr(idx) = ptr(idx) + 1;
        end
    end

    patches = mat2cell(patchElements, patchCounts);
end